import os
import json
import platform
from functools import lru_cache
from time import sleep
from urllib.parse import urlparse
import requests

from playwright._repo_version import version as PLAYWRIGHT_VERSION
from playwright.sync_api import Error as PlaywrightError, TimeoutError as PlaywrightTimeoutError
from percy.version import __version__ as SDK_VERSION
from percy.page_metadata import PageMetaData

# Collect client environment information
CLIENT_INFO = "percy-playwright-python/" + SDK_VERSION
ENV_INFO = ["playwright/" + PLAYWRIGHT_VERSION, "python/" + platform.python_version()]

def _get_bool_env(key):
    """Get boolean value from environment variable."""
    return os.environ.get(key, "").lower() == "true"

# Maybe get the CLI API address from the environment
PERCY_CLI_API = os.environ.get("PERCY_CLI_API") or "http://localhost:5338"
PERCY_DEBUG = os.environ.get("PERCY_LOGLEVEL") == "debug"
RESPONSIVE_CAPTURE_SLEEP_TIME = os.environ.get("RESPONSIVE_CAPTURE_SLEEP_TIME")
PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT = _get_bool_env("PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT")
PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE = _get_bool_env("PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE")

# for logging
LABEL = "[\u001b[35m" + ("percy:python" if PERCY_DEBUG else "percy") + "\u001b[39m]"

def log(message, lvl="info"):
    message = f"{LABEL} {message}"
    try:
        requests.post(
            f"{PERCY_CLI_API}/percy/log",
            json={"message": message, "level": lvl},
            timeout=5,
        )
    except Exception as e:
        if PERCY_DEBUG:
            print(f"Sending log to CLI Failed {e}")
    finally:
        # Only log if lvl is 'debug' and PERCY_DEBUG is True
        if lvl != "debug" or PERCY_DEBUG:
            print(message)


# Check if Percy is enabled, caching the result so it is only checked once
@lru_cache(maxsize=None)
def _is_percy_enabled():
    try:
        response = requests.get(f"{PERCY_CLI_API}/percy/healthcheck", timeout=30)
        response.raise_for_status()
        data = response.json()
        session_type = data.get("type", None)
        widths = data.get("widths", {})
        config = data.get("config", {})
        device_details = data.get("deviceDetails", [])

        if not data["success"]:
            raise Exception(data["error"])
        version = response.headers.get("x-percy-core-version")

        if not version:
            print(
                f"{LABEL} You may be using @percy/agent "
                "which is no longer supported by this SDK. "
                "Please uninstall @percy/agent and install @percy/cli instead. "
                "https://www.browserstack.com/docs/percy/migration/migrate-to-cli"
            )
            return False

        if version.split(".")[0] != "1":
            print(f"{LABEL} Unsupported Percy CLI version, {version}")
            return False

        return {
            "session_type": session_type,
            "config": config,
            "widths": widths,
            "device_details": device_details,
        }
    except Exception as e:
        print(f"{LABEL} Percy is not running, disabling snapshots")
        if PERCY_DEBUG:
            print(f"{LABEL} {e}")
        return False


# Fetch the @percy/dom script, caching the result so it is only fetched once
@lru_cache(maxsize=None)
def fetch_percy_dom():
    response = requests.get(f"{PERCY_CLI_API}/percy/dom.js", timeout=30)
    response.raise_for_status()
    return response.text


def process_frame(page, frame, options, percy_dom_script):
    """
    Processes a single cross-origin frame to capture its snapshot and resources.

    Args:
        page: The main page object
        frame: The frame to process
        options: Snapshot options
        percy_dom_script: The Percy DOM serialization script

    Returns:
        Dictionary containing iframe data, snapshot, and URL
    """
    frame_url = frame.url

    try:
        # Inject Percy DOM into the cross-origin frame
        frame.evaluate(percy_dom_script)

        # enableJavaScript=True prevents the standard iframe serialization logic from running.
        # This is necessary because we're manually handling cross-origin iframe serialization here.
        iframe_snapshot = frame.evaluate(
            f"PercyDOM.serialize({json.dumps({**options, 'enableJavaScript': True})})"
        )

        # Get the iframe's element data from the main page context
        iframe_data = page.evaluate(
            """(fUrl) => {
                const iframes = Array.from(document.querySelectorAll('iframe'));
                const matchingIframe = iframes.find(iframe => iframe.src.startsWith(fUrl));
                if (matchingIframe) {
                    return {
                        percyElementId: matchingIframe.getAttribute('data-percy-element-id')
                    };
                }
            }""",
            frame_url
        )

        if not iframe_data or not iframe_data.get("percyElementId"):
            log(
                f"Skipping cross-origin frame {frame_url}: "
                "no matching iframe element with percyElementId found on main page",
                "debug"
            )
            return None

        return {
            "iframeData": iframe_data,
            "iframeSnapshot": iframe_snapshot,
            "frameUrl": frame_url
        }
    except Exception as e:
        log(f"Failed to process cross-origin frame {frame_url}: {e}", "debug")
        return None


def get_serialized_dom(page, cookies, percy_dom_script=None, **kwargs):
    """
    Serializes the DOM and captures cross-origin iframes.

    Args:
        page: The page object
        cookies: Page cookies
        percy_dom_script: The Percy DOM serialization script
        **kwargs: Additional options

    Returns:
        Dictionary containing the DOM snapshot with cross-origin iframe data
    """
    dom_snapshot = page.evaluate(f"PercyDOM.serialize({json.dumps(kwargs)})")

    # Process CORS IFrames
    # Note: Blob URL handling (data-src images, blob background images) is now handled
    # in the CLI via async DOM serialization. This section only handles cross-origin
    # iframe serialization and resource merging.
    try:
        page_url = urlparse(page.url)
        frames = page.frames

        # Filter for cross-origin frames (excluding about:blank)
        cross_origin_frames = [
            frame for frame in frames
            if frame.url != "about:blank" and urlparse(frame.url).netloc != page_url.netloc
        ]

        if cross_origin_frames and percy_dom_script:
            processed_frames = []
            for frame in cross_origin_frames:
                result = process_frame(page, frame, kwargs, percy_dom_script)
                if result:
                    processed_frames.append(result)

            if processed_frames:
                dom_snapshot["corsIframes"] = processed_frames
    except Exception as e:
        log(f"Failed to process cross-origin iframes: {e}", "debug")

    dom_snapshot["cookies"] = cookies
    return dom_snapshot


# pylint: disable=too-many-arguments, too-many-branches
def create_region(
    boundingBox=None,
    elementXpath=None,
    elementCSS=None,
    padding=None,
    algorithm="ignore",
    diffSensitivity=None,
    imageIgnoreThreshold=None,
    carouselsEnabled=None,
    bannersEnabled=None,
    adsEnabled=None,
    diffIgnoreThreshold=None
    ):

    element_selector = {}
    if boundingBox:
        element_selector["boundingBox"] = boundingBox
    if elementXpath:
        element_selector["elementXpath"] = elementXpath
    if elementCSS:
        element_selector["elementCSS"] = elementCSS

    region = {
        "algorithm": algorithm,
        "elementSelector": element_selector
    }

    if padding:
        region["padding"] = padding

    configuration = {}
    if algorithm in ["standard", "intelliignore"]:
        if diffSensitivity is not None:
            configuration["diffSensitivity"] = diffSensitivity
        if imageIgnoreThreshold is not None:
            configuration["imageIgnoreThreshold"] = imageIgnoreThreshold
        if carouselsEnabled is not None:
            configuration["carouselsEnabled"] = carouselsEnabled
        if bannersEnabled is not None:
            configuration["bannersEnabled"] = bannersEnabled
        if adsEnabled is not None:
            configuration["adsEnabled"] = adsEnabled

    if configuration:
        region["configuration"] = configuration

    assertion = {}
    if diffIgnoreThreshold is not None:
        assertion["diffIgnoreThreshold"] = diffIgnoreThreshold

    if assertion:
        region["assertion"] = assertion

    return region



def calculate_default_height(current_height, config=None, **kwargs):
    """Calculate default height for responsive capture."""
    if not PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT:
        return current_height

    config_min_height = (config or {}).get("snapshot", {}).get("minHeight")
    min_height = kwargs.get("min_height") or config_min_height or current_height
    return min_height


def get_responsive_widths(widths=None):
    """Gets computed responsive widths from the Percy server for responsive snapshot capture."""
    if widths is None:
        widths = []
    try:
        # Ensure widths is a list
        widths_list = widths if isinstance(widths, list) else []
        query_param = f"?widths={','.join(map(str, widths_list))}" if widths_list else ""
        response = requests.get(
            f"{PERCY_CLI_API}/percy/widths-config{query_param}",
            timeout=30
        )
        response.raise_for_status()
        data = response.json()
        widths_data = data.get("widths")
        if not isinstance(widths_data, list):
            msg = "Update Percy CLI to the latest version to use responsiveSnapshotCapture"
            raise Exception(msg)
        return widths_data
    except Exception as e:
        log(f"Failed to get responsive widths: {e}.", "debug")
        msg = "Update Percy CLI to the latest version to use responsiveSnapshotCapture"
        raise Exception(msg) from e


def _responsive_sleep():
    """Sleep for the configured responsive capture sleep time if positive."""
    if not RESPONSIVE_CAPTURE_SLEEP_TIME:
        return
    try:
        if (secs := int(RESPONSIVE_CAPTURE_SLEEP_TIME)) > 0:
            sleep(secs)
    except (TypeError, ValueError):
        pass


def change_window_dimension_and_wait(page, width, height, resize_count):
    try:
        page.set_viewport_size({"width": width, "height": height})
    except PlaywrightError as e:
        log(f"Resizing viewport failed for width {width}: {e}", "debug")

    try:
        page.wait_for_function(
            f"window.resizeCount === {resize_count}", timeout=1000
        )
    except PlaywrightTimeoutError:
        log(f"Timed out waiting for window resize event for width {width}", "debug")


def capture_responsive_dom(page, cookies, percy_dom_script=None, config=None, **kwargs):
    viewport = page.viewport_size or page.evaluate(
        "() => ({ width: window.innerWidth, height: window.innerHeight })"
    )
    default_height = calculate_default_height(viewport["height"], config=config, **kwargs)

    # Get width and height combinations from CLI
    width_heights = get_responsive_widths(kwargs.get("widths", []))

    dom_snapshots = []
    last_window_width = viewport["width"]
    resize_count = 0
    page.evaluate("PercyDOM.waitForResize()")

    for width_height in width_heights:
        # Apply default height if not provided by CLI
        height = width_height.get("height") or default_height
        width = width_height["width"]
        if last_window_width != width:
            resize_count += 1
            change_window_dimension_and_wait(
                page, width, height, resize_count
            )
            last_window_width = width

        if PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE:
            page.reload()
            page.evaluate(percy_dom_script)
            page.evaluate("PercyDOM.waitForResize()")
            resize_count = 0

        _responsive_sleep()
        snapshot = get_serialized_dom(page, cookies, percy_dom_script, **kwargs)
        snapshot["width"] = width
        dom_snapshots.append(snapshot)

    change_window_dimension_and_wait(
        page, viewport["width"], viewport["height"], resize_count + 1
    )
    return dom_snapshots


def is_responsive_snapshot_capture(config, **kwargs):
    # Don't run responsive snapshot capture when defer uploads is enabled
    if "percy" in config and config["percy"].get("deferUploads", False):
        return False

    return (
        kwargs.get("responsive_snapshot_capture", False)
        or kwargs.get("responsiveSnapshotCapture", False)
        or (
            "snapshot" in config
            and config["snapshot"].get("responsiveSnapshotCapture")
        )
    )


# Take a DOM snapshot and post it to the snapshot endpoint
def percy_snapshot(page, name, **kwargs):
    data = _is_percy_enabled()
    if not data:
        return None

    if data["session_type"] == "automate":
        raise Exception(
            "Invalid function call - "
            "percy_snapshot(). "
            "Please use percy_screenshot() function while using Percy with Automate. "
            "For more information on usage of PercyScreenshot, "
            "refer https://www.browserstack.com/docs/percy/integrate/functional-and-visual"
        )

    try:
        # Inject the DOM serialization script
        percy_dom_script = fetch_percy_dom()
        page.evaluate(percy_dom_script)
        cookies = page.context.cookies()

        # Serialize and capture the DOM
        if is_responsive_snapshot_capture(data["config"], **kwargs):
            dom_snapshot = capture_responsive_dom(
                page, cookies, percy_dom_script, config=data["config"], **kwargs
            )
        else:
            dom_snapshot = get_serialized_dom(page, cookies, percy_dom_script, **kwargs)

        # Post the DOM to the snapshot endpoint with snapshot options and other info
        response = requests.post(
            f"{PERCY_CLI_API}/percy/snapshot",
            json={
                **kwargs,
                **{
                    "client_info": CLIENT_INFO,
                    "environment_info": ENV_INFO,
                    "dom_snapshot": dom_snapshot,
                    "url": page.url,
                    "name": name,
                },
            },
            timeout=600,
        )

        # Handle errors
        response.raise_for_status()
        response_data = response.json()

        if not response_data["success"]:
            raise Exception(response_data["error"])
        return response_data.get("data", None)
    except Exception as e:
        log(f'Could not take DOM snapshot "{name}"')
        log(f"{e}")
        return None


def percy_automate_screenshot(page, name, options=None, **kwargs):
    data = _is_percy_enabled()
    if not data:
        return None

    if data["session_type"] == "web":
        raise Exception(
            "Invalid function call - "
            "percy_screenshot(). Please use percy_snapshot() function for taking screenshot. "
            "percy_screenshot() should be used only while using Percy with Automate. "
            "For more information on usage of percy_snapshot(), "
            "refer doc for your language https://www.browserstack.com/docs/percy/integrate/overview"
        )

    if options is None:
        options = {}

    try:
        metadata = PageMetaData(page)

        # Post to automateScreenshot endpoint with page options and other info
        response = requests.post(
            f"{PERCY_CLI_API}/percy/automateScreenshot",
            json={
                **kwargs,
                **{
                    "client_info": CLIENT_INFO,
                    "environment_info": ENV_INFO,
                    "sessionId": metadata.automate_session_id,
                    "pageGuid": metadata.page_guid,
                    "frameGuid": metadata.frame_guid,
                    "framework": metadata.framework,
                    "snapshotName": name,
                    "options": options,
                },
            },
            timeout=600,
        )

        # Handle errors
        response.raise_for_status()
        response_data = response.json()

        if not response_data["success"]:
            raise Exception(response_data["error"])

        return response_data.get("data", None)
    except Exception as e:
        log(f'Could not take Screenshot "{name}"')
        log(f"{e}")
        return None



# pylint: disable=[abstract-class-instantiated, arguments-differ, protected-access, too-many-lines]
import json
import unittest
import platform
from threading import Thread
from http.server import BaseHTTPRequestHandler, HTTPServer
from unittest.mock import patch, MagicMock, call
import httpretty

from playwright.sync_api import sync_playwright
from playwright.sync_api import Error as PlaywrightError, TimeoutError as PlaywrightTimeoutError
from playwright._repo_version import version as PLAYWRIGHT_VERSION
from percy.version import __version__ as SDK_VERSION
from percy.screenshot import (
    _is_percy_enabled,
    fetch_percy_dom,
    percy_snapshot,
    percy_automate_screenshot,
    create_region,
    is_responsive_snapshot_capture,
    calculate_default_height,
    get_responsive_widths,
    capture_responsive_dom,
    change_window_dimension_and_wait,
    get_serialized_dom,
    process_frame,
    log
)
import percy.screenshot as local

LABEL = local.LABEL


# mock a simple webpage to snapshot
class MockServerRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-Type", "text/html")
        self.end_headers()
        self.wfile.write(("Snapshot Me").encode("utf-8"))

    def log_message(self, format, *args):
        return


# daemon threads automatically shut down when the main process exits
mock_server = HTTPServer(("localhost", 8000), MockServerRequestHandler)
mock_server_thread = Thread(target=mock_server.serve_forever)
mock_server_thread.daemon = True
mock_server_thread.start()

# initializing mock data
data_object = {"sync": "true", "diff": 0}


# mock helpers
# pylint: disable=too-many-arguments
def mock_healthcheck(
    fail=False, fail_how="error", session_type=None,
    widths=None, config=None, device_details=None
):
    health_body = {"success": True}
    health_headers = {"X-Percy-Core-Version": "1.0.0"}
    health_status = 200

    if fail and fail_how == "error":
        health_body = {"success": False, "error": "test"}
        health_status = 500
    elif fail and fail_how == "wrong-version":
        health_headers = {"X-Percy-Core-Version": "2.0.0"}
    elif fail and fail_how == "no-version":
        health_headers = {}

    if session_type:
        health_body["type"] = session_type
    if widths:
        health_body["widths"] = widths
    if config:
        health_body["config"] = config
    if device_details:
        health_body["deviceDetails"] = device_details

    health_body = json.dumps(health_body)
    httpretty.register_uri(
        httpretty.GET,
        "http://localhost:5338/percy/healthcheck",
        body=health_body,
        adding_headers=health_headers,
        status=health_status,
    )
    httpretty.register_uri(
        httpretty.GET,
        "http://localhost:5338/percy/dom.js",
        body=(
            "window.PercyDOM = { serialize: () => { return { html: "
            "document.documentElement.outerHTML } }, waitForResize: () => { "
            "if(!window.resizeCount) { window.addEventListener('resize', () => "
            "window.resizeCount++) } window.resizeCount = 0; }}"
        ),
        status=200,
    )


def mock_logger():
    httpretty.register_uri(
        httpretty.POST,
        "http://localhost:5338/percy/log",
        body=json.dumps({"success": True}),
        status=200,
    )


def mock_snapshot(fail=False, data=False):
    httpretty.register_uri(
        httpretty.POST,
        "http://localhost:5338/percy/snapshot",
        body=json.dumps(
            {
                "success": "false" if fail else "true",
                "error": "test" if fail else None,
                "data": data_object if data else None,
            }
        ),
        status=(500 if fail else 200),
    )


class TestPercySnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.p = sync_playwright().start()
        # Launch the browser
        cls.browser = cls.p.chromium.launch(
            headless=True
        )  # Set headless=True if you don't want to see the browser
        context = cls.browser.new_context()
        cls.page = context.new_page()

    @classmethod
    def tearDownClass(cls):
        cls.browser.close()
        cls.p.stop()

    def setUp(self):
        # clear the cached value for testing
        local._is_percy_enabled.cache_clear()
        local.fetch_percy_dom.cache_clear()
        self.page.goto("http://localhost:8000")
        httpretty.enable()

    def tearDown(self):
        httpretty.disable()
        httpretty.reset()

    def test_throws_error_when_a_page_is_not_provided(self):
        with self.assertRaises(Exception):
            percy_snapshot()

    def test_throws_error_when_a_name_is_not_provided(self):
        with self.assertRaises(Exception):
            percy_snapshot(self.page)

    def test_disables_snapshots_when_the_healthcheck_fails(self):
        mock_healthcheck(fail=True)

        with patch("builtins.print") as mock_print:
            percy_snapshot(self.page, "Snapshot 1")
            percy_snapshot(self.page, "Snapshot 2")

            mock_print.assert_called_with(
                f"{LABEL} Percy is not running, disabling snapshots"
            )

        self.assertEqual(httpretty.last_request().path, "/percy/healthcheck")

    def test_disables_snapshots_when_the_healthcheck_version_is_wrong(self):
        mock_healthcheck(fail=True, fail_how="wrong-version")

        with patch("builtins.print") as mock_print:
            percy_snapshot(self.page, "Snapshot 1")
            percy_snapshot(self.page, "Snapshot 2")

            mock_print.assert_called_with(
                f"{LABEL} Unsupported Percy CLI version, 2.0.0"
            )

        self.assertEqual(httpretty.last_request().path, "/percy/healthcheck")

    def test_disables_snapshots_when_the_healthcheck_version_is_missing(self):
        mock_healthcheck(fail=True, fail_how="no-version")

        with patch("builtins.print") as mock_print:
            percy_snapshot(self.page, "Snapshot 1")
            percy_snapshot(self.page, "Snapshot 2")

            mock_print.assert_called_with(
                f"{LABEL} You may be using @percy/agent which is no longer supported by this SDK. "
                "Please uninstall @percy/agent and install @percy/cli instead. "
                "https://www.browserstack.com/docs/percy/migration/migrate-to-cli"
            )

        self.assertEqual(httpretty.last_request().path, "/percy/healthcheck")

    def test_posts_snapshots_to_the_local_percy_server(self):
        mock_healthcheck()
        mock_snapshot()

        percy_snapshot(self.page, "Snapshot 1")
        response = percy_snapshot(self.page, "Snapshot 2", enable_javascript=True)

        self.assertEqual(httpretty.last_request().path, "/percy/snapshot")

        s1 = httpretty.latest_requests()[2].parsed_body
        self.assertEqual(s1["name"], "Snapshot 1")
        self.assertEqual(s1["url"], "http://localhost:8000/")
        self.assertEqual(
            s1["dom_snapshot"]["html"], "<html><head></head><body>Snapshot Me</body></html>"
        )
        self.assertIn("cookies", s1["dom_snapshot"])
        self.assertRegex(s1["client_info"], r"percy-playwright-python/\d+")
        self.assertRegex(s1["environment_info"][0], r"playwright/\d+")
        self.assertRegex(s1["environment_info"][1], r"python/\d+")

        s2 = httpretty.latest_requests()[3].parsed_body
        self.assertEqual(s2["name"], "Snapshot 2")
        self.assertEqual(s2["enable_javascript"], True)
        self.assertEqual(response, None)

    def test_posts_snapshots_to_the_local_percy_server_for_sync(self):
        mock_healthcheck()
        mock_snapshot(False, True)

        percy_snapshot(self.page, "Snapshot 1")
        response = percy_snapshot(
            self.page, "Snapshot 2", enable_javascript=True, sync=True
        )

        self.assertEqual(httpretty.last_request().path, "/percy/snapshot")

        s1 = httpretty.latest_requests()[2].parsed_body
        self.assertEqual(s1["name"], "Snapshot 1")
        self.assertEqual(s1["url"], "http://localhost:8000/")
        self.assertEqual(
            s1["dom_snapshot"]["html"], "<html><head></head><body>Snapshot Me</body></html>"
        )
        self.assertIn("cookies", s1["dom_snapshot"])
        self.assertRegex(s1["client_info"], r"percy-playwright-python/\d+")
        self.assertRegex(s1["environment_info"][0], r"playwright/\d+")
        self.assertRegex(s1["environment_info"][1], r"python/\d+")

        s2 = httpretty.latest_requests()[3].parsed_body
        self.assertEqual(s2["name"], "Snapshot 2")
        self.assertEqual(s2["enable_javascript"], True)
        self.assertEqual(s2["sync"], True)
        self.assertEqual(response, data_object)

        mock_healthcheck()
        mock_snapshot()

        percy_snapshot(self.page, "Snapshot")

        self.assertEqual(httpretty.last_request().path, "/percy/snapshot")

        s1 = httpretty.latest_requests()[-1].parsed_body
        self.assertEqual(s1["name"], "Snapshot")
        self.assertEqual(s1["url"], "http://localhost:8000/")
        self.assertEqual(
            s1["dom_snapshot"]["html"], "<html><head></head><body>Snapshot Me</body></html>"
        )
        self.assertIn("cookies", s1["dom_snapshot"])

    def test_handles_snapshot_errors(self):
        mock_healthcheck(session_type="web")
        mock_snapshot(fail=True)
        mock_logger()

        with patch("builtins.print") as mock_print:
            percy_snapshot(self.page, "Snapshot 1")

            mock_print.assert_any_call(
                f'{LABEL} Could not take DOM snapshot "Snapshot 1"'
            )

    def test_raise_error_poa_token_with_snapshot(self):
        mock_healthcheck(session_type="automate")

        with self.assertRaises(Exception) as context:
            percy_snapshot(self.page, "Snapshot 1")

        self.assertEqual(
            "Invalid function call - "
            "percy_snapshot(). Please use percy_screenshot() "
            "function while using Percy with Automate."
            " For more information on usage of PercyScreenshot, refer https://www.browserstack.com/"
            "docs/percy/integrate/functional-and-visual",
            str(context.exception),
        )


class TestPercyFunctions(unittest.TestCase):
    @patch("requests.get")
    def test_is_percy_enabled(self, mock_get):
        # Mock successful health check
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {
            "success": True,
            "type": "web",
            "config": {},
            "widths": {},
            "deviceDetails": [],
        }
        mock_get.return_value.headers = {"x-percy-core-version": "1.0.0"}

        self.assertEqual(
            _is_percy_enabled(),
            {
                "session_type": "web",
                "config": {},
                "widths": {},
                "device_details": [],
            },
        )

        # Clear the cache to test the unsuccessful scenario
        _is_percy_enabled.cache_clear()

        # Mock unsuccessful health check
        mock_get.return_value.json.return_value = {"success": False, "error": "error"}
        self.assertFalse(_is_percy_enabled())

    @patch("requests.get")
    def test_fetch_percy_dom(self, mock_get):
        # Mock successful fetch of dom.js
        fetch_percy_dom.cache_clear()
        mock_get.return_value.status_code = 200
        mock_get.return_value.text = "some_js_code"

        self.assertEqual(fetch_percy_dom(), "some_js_code")

    @patch("requests.get")
    def test_fetch_percy_dom_raises(self, mock_get):
        fetch_percy_dom.cache_clear()
        mock_get.return_value.raise_for_status.side_effect = Exception("boom")

        with self.assertRaises(Exception):
            fetch_percy_dom()

    @patch("requests.post")
    @patch("percy.screenshot.fetch_percy_dom")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_snapshot(
        self, mock_is_percy_enabled, mock_fetch_percy_dom, mock_post
    ):
        # Mock Percy enabled
        mock_is_percy_enabled.return_value = {
            "session_type": "web",
            "config": {},
            "widths": {},
            "device_details": [],
        }
        mock_fetch_percy_dom.return_value = "some_js_code"
        page = MagicMock()
        page.evaluate.side_effect = [None, {"html": "<html></html>"}]
        page.context.cookies.return_value = []
        page.url = "http://example.com"
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "success": True,
            "data": "snapshot_data",
        }

        # Call the function
        result = percy_snapshot(page, "snapshot_name")

        # Check the results
        self.assertEqual(result, "snapshot_data")
        mock_post.assert_called_once()

    @patch("requests.post")
    @patch("percy.screenshot.fetch_percy_dom")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_snapshot_includes_cookies(
        self, mock_is_percy_enabled, mock_fetch_percy_dom, mock_post
    ):
        mock_is_percy_enabled.return_value = {
            "session_type": "web",
            "config": {},
            "widths": {},
            "device_details": [],
        }
        mock_fetch_percy_dom.return_value = "some_js_code"
        page = MagicMock()
        page.evaluate.side_effect = [None, {"html": "<html></html>"}]
        page.context.cookies.return_value = [{"name": "foo", "value": "bar"}]
        page.url = "http://example.com"
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "success": True,
            "data": "snapshot_data",
        }

        percy_snapshot(page, "snapshot_name")

        posted = mock_post.call_args.kwargs["json"]
        self.assertEqual(
            posted["dom_snapshot"]["cookies"], [{"name": "foo", "value": "bar"}]
        )

    def test_process_frame_returns_cors_iframe_data(self):
        page = MagicMock()
        page.evaluate.return_value = {"percyElementId": "iframe-1"}

        frame = MagicMock()
        frame.url = "http://cross-origin.example/frame"
        frame.evaluate.side_effect = [None, {"html": "<iframe></iframe>"}]

        result = process_frame(
            page,
            frame,
            {"enableJavaScript": False},
            "percy-dom-script"
        )

        self.assertEqual(
            result,
            {
                "iframeData": {"percyElementId": "iframe-1"},
                "iframeSnapshot": {"html": "<iframe></iframe>"},
                "frameUrl": "http://cross-origin.example/frame",
            },
        )

    def test_get_serialized_dom_adds_cors_iframes(self):
        page = MagicMock()
        page.url = "http://example.com"
        page.evaluate.return_value = {"html": "<html></html>"}

        same_origin_frame = MagicMock()
        same_origin_frame.url = "http://example.com/frame"
        cross_origin_frame = MagicMock()
        cross_origin_frame.url = "http://other.example/frame"
        page.frames = [same_origin_frame, cross_origin_frame]

        with patch("percy.screenshot.process_frame") as mock_process:
            mock_process.return_value = {"frameUrl": "http://other.example/frame"}
            dom_snapshot = get_serialized_dom(
                page,
                [{"name": "foo", "value": "bar"}],
                percy_dom_script="percy-dom"
            )

        mock_process.assert_called_once_with(page, cross_origin_frame, {}, "percy-dom")
        self.assertEqual(
            dom_snapshot["corsIframes"],
            [{"frameUrl": "http://other.example/frame"}]
        )
        self.assertEqual(
            dom_snapshot["cookies"],
            [{"name": "foo", "value": "bar"}]
        )

    def test_process_frame_returns_none_on_error(self):
        page = MagicMock()
        frame = MagicMock()
        frame.url = "http://cross-origin.example/frame"
        frame.evaluate.side_effect = Exception("boom")

        with patch("percy.screenshot.log") as mock_log:
            result = process_frame(page, frame, {}, "percy-dom-script")

        self.assertIsNone(result)
        mock_log.assert_called_once()

    def test_process_frame_returns_none_when_iframe_data_missing(self):
        page = MagicMock()
        page.evaluate.return_value = None  # iframe element not found on main page

        frame = MagicMock()
        frame.url = "http://cross-origin.example/frame"
        frame.evaluate.side_effect = [None, {"html": "<iframe></iframe>"}]

        with patch("percy.screenshot.log") as mock_log:
            result = process_frame(page, frame, {}, "percy-dom-script")

        self.assertIsNone(result)
        mock_log.assert_called_once_with(
            "Skipping cross-origin frame http://cross-origin.example/frame: "
            "no matching iframe element with percyElementId found on main page",
            "debug"
        )

    def test_process_frame_returns_none_when_percy_element_id_missing(self):
        page = MagicMock()
        page.evaluate.return_value = {}  # iframe found but lacks percyElementId

        frame = MagicMock()
        frame.url = "http://cross-origin.example/frame"
        frame.evaluate.side_effect = [None, {"html": "<iframe></iframe>"}]

        with patch("percy.screenshot.log") as mock_log:
            result = process_frame(page, frame, {}, "percy-dom-script")

        self.assertIsNone(result)
        mock_log.assert_called_once_with(
            "Skipping cross-origin frame http://cross-origin.example/frame: "
            "no matching iframe element with percyElementId found on main page",
            "debug"
        )

    def test_get_serialized_dom_skips_empty_cors_results(self):
        page = MagicMock()
        page.url = "http://example.com"
        page.evaluate.return_value = {"html": "<html></html>"}
        cross_origin_frame = MagicMock()
        cross_origin_frame.url = "http://other.example/frame"
        page.frames = [cross_origin_frame]

        with patch("percy.screenshot.process_frame") as mock_process:
            mock_process.return_value = None
            dom_snapshot = get_serialized_dom(
                page,
                [{"name": "foo", "value": "bar"}],
                percy_dom_script="percy-dom"
            )

        self.assertNotIn("corsIframes", dom_snapshot)
        self.assertEqual(
            dom_snapshot["cookies"],
            [{"name": "foo", "value": "bar"}]
        )

    def test_get_serialized_dom_logs_when_frame_processing_fails(self):
        class BadUrl:  # pylint: disable=too-few-public-methods
            def __str__(self):
                raise Exception("boom")

        page = MagicMock()
        page.url = BadUrl()
        page.evaluate.return_value = {"html": "<html></html>"}

        with patch("percy.screenshot.log") as mock_log:
            dom_snapshot = get_serialized_dom(
                page,
                [{"name": "foo", "value": "bar"}],
                percy_dom_script="percy-dom"
            )

        mock_log.assert_called_once()
        self.assertEqual(
            dom_snapshot["cookies"],
            [{"name": "foo", "value": "bar"}]
        )

    @patch("requests.post")
    @patch("percy.screenshot.capture_responsive_dom")
    @patch("percy.screenshot.fetch_percy_dom")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_snapshot_responsive_capture(
        self,
        mock_is_percy_enabled,
        mock_fetch_percy_dom,
        mock_capture_responsive_dom,
        mock_post,
    ):
        mock_is_percy_enabled.return_value = {
            "session_type": "web",
            "config": {"snapshot": {"responsiveSnapshotCapture": True}},
        }
        mock_fetch_percy_dom.return_value = "some_js_code"
        mock_capture_responsive_dom.return_value = [
            {"html": "<html></html>", "width": 375},
            {"html": "<html></html>", "width": 1280},
        ]
        page = MagicMock()
        page.evaluate.return_value = None
        page.context.cookies.return_value = [{"name": "foo", "value": "bar"}]
        page.url = "http://example.com"
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "success": True,
            "data": "snapshot_data",
        }

        percy_snapshot(page, "snapshot_name")

        mock_capture_responsive_dom.assert_called_once_with(
            page,
            [{"name": "foo", "value": "bar"}],
            "some_js_code",
            config={"snapshot": {"responsiveSnapshotCapture": True}},
        )
        posted = mock_post.call_args.kwargs["json"]
        self.assertEqual(posted["dom_snapshot"], mock_capture_responsive_dom.return_value)

    @patch("requests.post")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_automate_screenshot(self, mock_is_percy_enabled, mock_post):
        # Mock Percy enabled for automate
        _is_percy_enabled.cache_clear()
        mock_is_percy_enabled.return_value = {
            "session_type": "automate",
            "config": {},
        }
        page = MagicMock()

        page._impl_obj._guid = "page@abc"
        page.main_frame._impl_obj._guid = "frame@abc"
        page.context.browser._impl_obj._guid = "browser@abc"
        page.evaluate.return_value = '{"hashed_id": "session_id"}'

        # Mock the response for the POST request
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "success": True,
            "data": "screenshot_data",
        }

        # Call the function
        result = percy_automate_screenshot(page, "screenshot_name")

        # Assertions
        self.assertEqual(result, "screenshot_data")
        mock_post.assert_called_once_with(
            "http://localhost:5338/percy/automateScreenshot",
            json={
                "client_info": f"percy-playwright-python/{SDK_VERSION}",
                "environment_info": [
                    f"playwright/{PLAYWRIGHT_VERSION}",
                    f"python/{platform.python_version()}",
                ],
                "sessionId": "session_id",
                "pageGuid": "page@abc",
                "frameGuid": "frame@abc",
                "framework": "playwright",
                "snapshotName": "screenshot_name",
                "options": {},
            },
            timeout=600,
        )

    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_automate_screenshot_percy_disabled(self, mock_is_percy_enabled):
        """Test percy_automate_screenshot when Percy is not enabled."""
        _is_percy_enabled.cache_clear()
        mock_is_percy_enabled.return_value = False

        page = MagicMock()
        result = percy_automate_screenshot(page, "screenshot_name")

        # Should return None when Percy is disabled
        self.assertIsNone(result)

    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_automate_screenshot_invalid_call(self, mock_is_percy_enabled):
        # Mock Percy enabled for web
        mock_is_percy_enabled.return_value = {
            "session_type": "web",
            "config": {},
        }
        page = MagicMock()

        # Call the function and expect an exception
        with self.assertRaises(Exception) as context:
            percy_automate_screenshot(page, "screenshot_name")

        self.assertTrue("Invalid function call" in str(context.exception))

    @patch("requests.post")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_automate_screenshot_with_options(self, mock_is_percy_enabled, mock_post):
        """Test percy_automate_screenshot when options are provided (not None)."""
        # Mock Percy enabled for automate
        _is_percy_enabled.cache_clear()
        mock_is_percy_enabled.return_value = {
            "session_type": "automate",
            "config": {},
        }
        page = MagicMock()

        page._impl_obj._guid = "page@xyz"
        page.main_frame._impl_obj._guid = "frame@xyz"
        page.context.browser._impl_obj._guid = "browser@xyz"
        page.evaluate.return_value = '{"hashed_id": "session_xyz"}'

        # Mock the response for the POST request
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "success": True,
            "data": "screenshot_with_options",
        }

        # Call the function with explicit options (not None)
        test_options = {"freeze_animated_image": True, "freeze_image_by_selectors": [".animated"]}
        result = percy_automate_screenshot(page, "screenshot_name", options=test_options)

        # Assertions
        self.assertEqual(result, "screenshot_with_options")
        call_args = mock_post.call_args
        self.assertEqual(call_args.kwargs["json"]["options"], test_options)


class TestScreenshotEdgeCases(unittest.TestCase):
    def setUp(self):
        _is_percy_enabled.cache_clear()
        fetch_percy_dom.cache_clear()

    @patch("requests.get")
    def test_is_percy_enabled_request_error(self, mock_get):
        mock_get.side_effect = Exception("boom")
        with patch.object(local, "PERCY_DEBUG", True), patch("builtins.print") as mock_print:
            self.assertFalse(_is_percy_enabled())
            mock_print.assert_any_call(
                f"{LABEL} Percy is not running, disabling snapshots"
            )

    @patch("requests.get")
    def test_is_percy_enabled_raise_for_status(self, mock_get):
        mock_get.return_value.raise_for_status.side_effect = Exception("boom")
        mock_get.return_value.json.return_value = {"success": True}
        with patch.object(local, "PERCY_DEBUG", False):
            self.assertFalse(_is_percy_enabled())

    def test_log_debug_when_enabled(self):
        with patch("requests.post", side_effect=Exception("boom")), patch.object(
            local, "PERCY_DEBUG", True
        ), patch("builtins.print") as mock_print:
            log("message", lvl="debug")
            mock_print.assert_any_call("Sending log to CLI Failed boom")
            mock_print.assert_any_call(f"{LABEL} message")

    def test_log_debug_when_disabled(self):
        with patch("requests.post") as mock_post, patch.object(
            local, "PERCY_DEBUG", False
        ), patch("builtins.print") as mock_print:
            mock_post.return_value.status_code = 200
            log("message", lvl="debug")
            mock_print.assert_not_called()

    def test_log_info_level(self):
        """Test log function with info level (should always print)."""
        with patch("requests.post") as mock_post, patch.object(
            local, "PERCY_DEBUG", False
        ), patch("builtins.print") as mock_print:
            mock_post.return_value.status_code = 200
            log("info message", lvl="info")
            mock_print.assert_called_once_with(f"{LABEL} info message")

    def test_log_exception_debug_disabled(self):
        """Test log function when exception occurs and PERCY_DEBUG is False."""
        with patch("requests.post", side_effect=Exception("connection error")), patch.object(
            local, "PERCY_DEBUG", False
        ), patch("builtins.print") as mock_print:
            log("message with error", lvl="info")
            # Should still print the message in finally block, but not the exception
            mock_print.assert_called_once_with(f"{LABEL} message with error")

    def test_change_window_dimension_and_wait_errors(self):
        page = MagicMock()
        page.set_viewport_size.side_effect = PlaywrightError("boom")
        page.wait_for_function.side_effect = PlaywrightTimeoutError("boom")

        with patch("percy.screenshot.log") as mock_log:
            change_window_dimension_and_wait(page, 100, 200, 1)

        self.assertEqual(mock_log.call_count, 2)

    def test_capture_responsive_dom_no_reload_or_sleep(self):
        page = MagicMock()
        page.viewport_size = {"width": 800, "height": 600}
        page.evaluate = MagicMock()
        page.reload = MagicMock()

        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE", None), patch.object(
            local, "RESPONSIVE_CAPTURE_SLEEP_TIME", None
        ), patch(
            "percy.screenshot.get_responsive_widths"
        ) as mock_widths, patch(
            "percy.screenshot.get_serialized_dom"
        ) as mock_serialized, patch(
            "percy.screenshot.change_window_dimension_and_wait"
        ) as mock_resize:
            mock_widths.return_value = [{"width": 800, "height": 600}]
            mock_serialized.return_value = {"html": "<html></html>"}

            capture_responsive_dom(page, [{"name": "foo", "value": "bar"}])

        page.reload.assert_not_called()
        page.evaluate.assert_any_call("PercyDOM.waitForResize()")
        mock_resize.assert_called_once_with(page, 800, 600, 1)

    def test_capture_responsive_dom_none_viewport_falls_back_to_js(self):
        page = MagicMock()
        page.viewport_size = None
        page.evaluate = MagicMock(
            side_effect=lambda expr: {"width": 1024, "height": 768}
            if "innerWidth" in expr
            else None
        )
        page.reload = MagicMock()

        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE", None), patch.object(
            local, "RESPONSIVE_CAPTURE_SLEEP_TIME", None
        ), patch(
            "percy.screenshot.get_responsive_widths"
        ) as mock_widths, patch(
            "percy.screenshot.get_serialized_dom"
        ) as mock_serialized, patch(
            "percy.screenshot.change_window_dimension_and_wait"
        ) as mock_resize:
            mock_widths.return_value = [{"width": 1024, "height": 768}]
            mock_serialized.return_value = {"html": "<html></html>"}

            capture_responsive_dom(page, [{"name": "foo", "value": "bar"}])

        page.evaluate.assert_any_call(
            "() => ({ width: window.innerWidth, height: window.innerHeight })"
        )
        mock_resize.assert_called_once_with(page, 1024, 768, 1)

    @patch("requests.post")
    @patch("percy.screenshot.fetch_percy_dom")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_snapshot_response_error(
        self, mock_is_percy_enabled, mock_fetch_percy_dom, mock_post
    ):
        mock_is_percy_enabled.return_value = {
            "session_type": "web",
            "config": {},
        }
        mock_fetch_percy_dom.return_value = "some_js_code"
        page = MagicMock()
        page.evaluate.side_effect = [None, {"html": "<html></html>"}]
        page.context.cookies.return_value = []
        page.url = "http://example.com"
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "success": False,
            "error": "bad",
        }

        with patch("percy.screenshot.log") as mock_log:
            result = percy_snapshot(page, "snapshot_name")

        self.assertIsNone(result)
        mock_log.assert_any_call('Could not take DOM snapshot "snapshot_name"')

    @patch("requests.post")
    @patch("percy.screenshot.fetch_percy_dom")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_snapshot_request_error(
        self, mock_is_percy_enabled, mock_fetch_percy_dom, mock_post
    ):
        mock_is_percy_enabled.return_value = {
            "session_type": "web",
            "config": {},
        }
        mock_fetch_percy_dom.return_value = "some_js_code"
        page = MagicMock()
        page.evaluate.side_effect = [None, {"html": "<html></html>"}]
        page.context.cookies.return_value = []
        page.url = "http://example.com"
        mock_post.side_effect = Exception("boom")

        with patch("percy.screenshot.log") as mock_log:
            result = percy_snapshot(page, "snapshot_name")

        self.assertIsNone(result)
        mock_log.assert_any_call('Could not take DOM snapshot "snapshot_name"')

    @patch("requests.post")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_automate_screenshot_response_error(
        self, mock_is_percy_enabled, mock_post
    ):
        mock_is_percy_enabled.return_value = {
            "session_type": "automate",
            "config": {},
        }
        page = MagicMock()
        page._impl_obj._guid = "page@abc"
        page.main_frame._impl_obj._guid = "frame@abc"
        page.context.browser._impl_obj._guid = "browser@abc"
        page.evaluate.return_value = '{"hashed_id": "session_id"}'
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {
            "success": False,
            "error": "bad",
        }

        with patch("percy.screenshot.log") as mock_log:
            result = percy_automate_screenshot(page, "screenshot_name")

        self.assertIsNone(result)
        mock_log.assert_any_call('Could not take Screenshot "screenshot_name"')

    @patch("requests.post")
    @patch("percy.screenshot._is_percy_enabled")
    def test_percy_automate_screenshot_request_error(
        self, mock_is_percy_enabled, mock_post
    ):
        mock_is_percy_enabled.return_value = {
            "session_type": "automate",
            "config": {},
        }
        page = MagicMock()
        page._impl_obj._guid = "page@abc"
        page.main_frame._impl_obj._guid = "frame@abc"
        page.context.browser._impl_obj._guid = "browser@abc"
        page.evaluate.return_value = '{"hashed_id": "session_id"}'
        mock_post.side_effect = Exception("boom")

        with patch("percy.screenshot.log") as mock_log:
            result = percy_automate_screenshot(page, "screenshot_name")

        self.assertIsNone(result)
        mock_log.assert_any_call('Could not take Screenshot "screenshot_name"')

class TestCreateRegion(unittest.TestCase):

    def test_create_region_with_all_params(self):
        result = create_region(
            boundingBox={"x": 10, "y": 20, "width": 100, "height": 200},
            elementXpath="//*[@id='test']",
            elementCSS=".test-class",
            padding=10,
            algorithm="intelliignore",
            diffSensitivity=0.8,
            imageIgnoreThreshold=0.5,
            carouselsEnabled=True,
            bannersEnabled=False,
            adsEnabled=True,
            diffIgnoreThreshold=0.2
        )

        expected_result = {
            "algorithm": "intelliignore",
            "elementSelector": {
                "boundingBox": {"x": 10, "y": 20, "width": 100, "height": 200},
                "elementXpath": "//*[@id='test']",
                "elementCSS": ".test-class"
            },
            "padding": 10,
            "configuration": {
                "diffSensitivity": 0.8,
                "imageIgnoreThreshold": 0.5,
                "carouselsEnabled": True,
                "bannersEnabled": False,
                "adsEnabled": True
            },
            "assertion": {
                "diffIgnoreThreshold": 0.2
            }
        }

        self.assertEqual(result, expected_result)


# pylint: disable=too-many-public-methods
class TestResponsiveHelpers(unittest.TestCase):
    def test_is_responsive_snapshot_capture_from_kwargs(self):
        self.assertTrue(is_responsive_snapshot_capture({}, responsive_snapshot_capture=True))
        self.assertTrue(is_responsive_snapshot_capture({}, responsiveSnapshotCapture=True))

    def test_is_responsive_snapshot_capture_from_config(self):
        config = {"snapshot": {"responsiveSnapshotCapture": True}}
        self.assertTrue(is_responsive_snapshot_capture(config))

    def test_is_responsive_snapshot_capture_defer_uploads(self):
        config = {
            "percy": {"deferUploads": True},
            "snapshot": {"responsiveSnapshotCapture": True},
        }
        self.assertFalse(is_responsive_snapshot_capture(config, responsiveSnapshotCapture=True))

    def test_calculate_default_height_env_disabled(self):
        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT", None):
            self.assertEqual(calculate_default_height(123), 123)

    def test_calculate_default_height_env_enabled(self):
        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT", "1"):
            self.assertEqual(calculate_default_height(123, min_height=200), 200)

    def test_calculate_default_height_env_enabled_handles_error(self):
        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT", "1"):
            self.assertEqual(calculate_default_height(321), 321)

    def test_calculate_default_height_uses_config_min_height(self):
        config = {"snapshot": {"minHeight": 500}}
        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT", "1"):
            self.assertEqual(calculate_default_height(321, config=config), 500)

    def test_calculate_default_height_uses_current_height_as_fallback(self):
        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_MIN_HEIGHT", "1"):
            self.assertEqual(calculate_default_height(321), 321)

    @patch("requests.get")
    def test_get_responsive_widths(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {
            "widths": [
                {"width": 375, "height": 667},
                {"width": 1280, "height": 900},
            ]
        }

        widths = get_responsive_widths([375, 1280])

        self.assertEqual(
            widths,
            [
                {"width": 375, "height": 667},
                {"width": 1280, "height": 900},
            ],
        )
        mock_get.assert_called_once_with(
            "http://localhost:5338/percy/widths-config?widths=375,1280",
            timeout=30,
        )

    @patch("requests.get")
    def test_get_responsive_widths_missing_widths(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {"widths": "not-a-list"}

        with self.assertRaises(Exception) as context:
            get_responsive_widths([800])

        self.assertEqual(
            str(context.exception),
            "Update Percy CLI to the latest version to use responsiveSnapshotCapture",
        )

    @patch("requests.get")
    def test_get_responsive_widths_with_none(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {
            "widths": [
                {"width": 375, "height": 667},
            ]
        }

        widths = get_responsive_widths(None)

        self.assertEqual(
            widths,
            [
                {"width": 375, "height": 667},
            ],
        )
        mock_get.assert_called_once_with(
            "http://localhost:5338/percy/widths-config",
            timeout=30,
        )

    @patch("requests.get")
    def test_get_responsive_widths_with_empty_list(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {
            "widths": [
                {"width": 375, "height": 667},
            ]
        }

        widths = get_responsive_widths([])

        self.assertEqual(
            widths,
            [
                {"width": 375, "height": 667},
            ],
        )
        mock_get.assert_called_once_with(
            "http://localhost:5338/percy/widths-config",
            timeout=30,
        )

    def test_capture_responsive_dom_calls_resize_reload_sleep(self):
        page = MagicMock()
        page.viewport_size = {"width": 800, "height": 600}
        page.evaluate = MagicMock()
        page.reload = MagicMock()

        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE", "1"), patch.object(
            local, "RESPONSIVE_CAPTURE_SLEEP_TIME", "1"
        ), patch(
            "percy.screenshot.get_responsive_widths"
        ) as mock_widths, patch(
            "percy.screenshot.get_serialized_dom"
        ) as mock_serialized, patch(
            "percy.screenshot.change_window_dimension_and_wait"
        ) as mock_resize, patch(
            "percy.screenshot.fetch_percy_dom"
        ) as mock_fetch, patch(
            "percy.screenshot.sleep"
        ) as mock_sleep:
            mock_widths.return_value = [
                {"width": 800, "height": 600},
                {"width": 1200, "height": 700},
            ]
            mock_serialized.side_effect = [
                {"html": "<html></html>"},
                {"html": "<html></html>"},
            ]
            mock_fetch.return_value = "dom-script"

            result = capture_responsive_dom(page, [{"name": "foo", "value": "bar"}], "dom-script")

        page.evaluate.assert_any_call("PercyDOM.waitForResize()")
        page.evaluate.assert_any_call("dom-script")
        self.assertEqual(page.evaluate.call_count, 5)
        self.assertEqual(page.reload.call_count, 2)
        mock_sleep.assert_any_call(1)
        self.assertEqual(mock_sleep.call_count, 2)
        mock_resize.assert_has_calls(
            [
                call(page, 1200, 700, 1),
                call(page, 800, 600, 1),
            ]
        )
        self.assertEqual(result[0]["width"], 800)
        self.assertEqual(result[1]["width"], 1200)

    def test_capture_responsive_dom_invalid_sleep_time(self):
        page = MagicMock()
        page.viewport_size = {"width": 800, "height": 600}
        page.evaluate = MagicMock()
        page.reload = MagicMock()

        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE", None), patch.object(
            local, "RESPONSIVE_CAPTURE_SLEEP_TIME", "not-a-number"
        ), patch(
            "percy.screenshot.get_responsive_widths"
        ) as mock_widths, patch(
            "percy.screenshot.get_serialized_dom"
        ) as mock_serialized, patch(
            "percy.screenshot.change_window_dimension_and_wait"
        ), patch(
            "percy.screenshot.sleep"
        ) as mock_sleep:
            mock_widths.return_value = [{"width": 800, "height": 600}]
            mock_serialized.return_value = {"html": "<html></html>"}

            capture_responsive_dom(page, [])

        mock_sleep.assert_not_called()

    def test_capture_responsive_dom_zero_sleep_time(self):
        page = MagicMock()
        page.viewport_size = {"width": 800, "height": 600}
        page.evaluate = MagicMock()

        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE", None), patch.object(
            local, "RESPONSIVE_CAPTURE_SLEEP_TIME", "0"
        ), patch(
            "percy.screenshot.get_responsive_widths"
        ) as mock_widths, patch(
            "percy.screenshot.get_serialized_dom"
        ) as mock_serialized, patch(
            "percy.screenshot.change_window_dimension_and_wait"
        ), patch(
            "percy.screenshot.sleep"
        ) as mock_sleep:
            mock_widths.return_value = [{"width": 800, "height": 600}]
            mock_serialized.return_value = {"html": "<html></html>"}

            capture_responsive_dom(page, [])

        mock_sleep.assert_not_called()

    def test_capture_responsive_dom_skips_resize_for_same_width(self):
        """Test that resize is skipped when consecutive widths are the same."""
        page = MagicMock()
        page.viewport_size = {"width": 800, "height": 600}
        page.evaluate = MagicMock()

        with patch.object(local, "PERCY_RESPONSIVE_CAPTURE_RELOAD_PAGE", None), patch.object(
            local, "RESPONSIVE_CAPTURE_SLEEP_TIME", None
        ), patch(
            "percy.screenshot.get_responsive_widths"
        ) as mock_widths, patch(
            "percy.screenshot.get_serialized_dom"
        ) as mock_serialized, patch(
            "percy.screenshot.change_window_dimension_and_wait"
        ) as mock_resize:
            # Test with duplicate widths: 800 (same as viewport), 1200, 1200 (duplicate), 1400
            mock_widths.return_value = [
                {"width": 800, "height": 600},
                {"width": 1200, "height": 700},
                {"width": 1200, "height": 700},  # Duplicate - should not trigger resize
                {"width": 1400, "height": 800},
            ]
            mock_serialized.side_effect = [
                {"html": "<html></html>"},
                {"html": "<html></html>"},
                {"html": "<html></html>"},
                {"html": "<html></html>"},
            ]

            result = capture_responsive_dom(page, [])

        # Verify resize is called only when width changes:
        # 1. First width (800) matches viewport - no resize
        # 2. Second width (1200) differs - resize to 1200
        # 3. Third width (1200) same as previous - no resize
        # 4. Fourth width (1400) differs - resize to 1400
        # 5. Final restore to viewport (800) - resize to 800
        mock_resize.assert_has_calls(
            [
                call(page, 1200, 700, 1),  # First change from 800 to 1200
                call(page, 1400, 800, 2),  # Second change from 1200 to 1400
                call(page, 800, 600, 3),   # Final restore
            ]
        )
        self.assertEqual(mock_resize.call_count, 3)
        self.assertEqual(len(result), 4)  # All 4 widths should have snapshots

    def test_create_region_with_minimal_params(self):
        result = create_region(
            algorithm="standard",
            boundingBox={"x": 10, "y": 20, "width": 100, "height": 200}
        )

        expected_result = {
            "algorithm": "standard",
            "elementSelector": {
                "boundingBox": {"x": 10, "y": 20, "width": 100, "height": 200}
            }
        }

        self.assertEqual(result, expected_result)

    def test_create_region_with_padding(self):
        result = create_region(
            algorithm="ignore",
            padding=15
        )

        expected_result = {
            "algorithm": "ignore",
            "elementSelector": {},
            "padding": 15
        }

        self.assertEqual(result, expected_result)

    def test_create_region_with_configuration_only_for_valid_algorithms(self):
        result = create_region(
            algorithm="intelliignore",
            diffSensitivity=0.9,
            imageIgnoreThreshold=0.7
        )

        expected_result = {
            "algorithm": "intelliignore",
            "elementSelector": {},
            "configuration": {
                "diffSensitivity": 0.9,
                "imageIgnoreThreshold": 0.7
            }
        }

        self.assertEqual(result, expected_result)

    def test_create_region_with_diffIgnoreThreshold_in_assertion(self):
        result = create_region(
            algorithm="standard",
            diffIgnoreThreshold=0.3
        )

        expected_result = {
            "algorithm": "standard",
            "elementSelector": {},
            "assertion": {
                "diffIgnoreThreshold": 0.3
            }
        }

        self.assertEqual(result, expected_result)

    def test_create_region_with_invalid_algorithm(self):
        result = create_region(
            algorithm="invalid_algorithm"
        )

        expected_result = {
            "algorithm": "invalid_algorithm",
            "elementSelector": {}
        }

        self.assertEqual(result, expected_result)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=[abstract-class-instantiated, arguments-differ, protected-access]
import unittest
from unittest.mock import patch, MagicMock
import percy


class TestPercyInit(unittest.TestCase):
    @patch("percy.percy_snapshot")
    def test_percySnapshot_wrapper(self, mock_percy_snapshot):
        """Test the percySnapshot backwards compatibility wrapper."""
        mock_page = MagicMock()
        mock_percy_snapshot.return_value = "snapshot_result"

        # Call via percy module to test the wrapper
        # percySnapshot(browser, *a, **kw) calls percy_snapshot(page=browser, *a, **kw)
        result = percy.percySnapshot(mock_page, "test_name", some_option=True)

        # Verify percy_snapshot was called correctly
        # Check that the function was called
        self.assertTrue(mock_percy_snapshot.called)
        # Check page argument and keyword argument
        actual_call = mock_percy_snapshot.call_args
        self.assertEqual(actual_call.kwargs['page'], mock_page)
        self.assertEqual(actual_call.kwargs['some_option'], True)
        self.assertEqual(result, "snapshot_result")

    @patch("percy.percy_automate_screenshot")
    def test_percy_screenshot_wrapper(self, mock_automate_screenshot):
        """Test the percy_screenshot wrapper."""
        mock_page = MagicMock()
        mock_automate_screenshot.return_value = "screenshot_result"

        # Call via percy module to test the wrapper
        result = percy.percy_screenshot(mock_page, "test_name", options={"key": "value"})

        # Verify percy_automate_screenshot was called correctly
        mock_automate_screenshot.assert_called_once_with(
            mock_page, "test_name", options={"key": "value"}
        )
        self.assertEqual(result, "screenshot_result")

    def test_version_imported(self):
        """Test that __version__ is properly imported."""
        self.assertIsNotNone(percy.__version__)
        self.assertIsInstance(percy.__version__, str)


if __name__ == "__main__":
    unittest.main()



# . # pylint: disable=[arguments-differ, protected-access]
import time
import unittest
from unittest.mock import patch
from percy.cache import Cache


class TestCache(unittest.TestCase):
    def setUp(self) -> None:
        self.cache = Cache
        self.session_id = "session_id_123"
        self.session_details = {
            "browser": "chrome",
            "platform": "windows",
            "browserVersion": "115.0.1",
            "hashed_id": "abcdef",
        }

        self.cache.set_cache(
            self.session_id, Cache.session_details, self.session_details
        )
        self.cache.set_cache(self.session_id, "key-1", "some-value")

    def test_set_cache(self):
        with self.assertRaises(Exception) as e:
            self.cache.set_cache(123, 123, 123)
        self.assertEqual(str(e.exception), "Argument session_id should be string")

        with self.assertRaises(Exception) as e:
            self.cache.set_cache(self.session_id, 123, 123)
        self.assertEqual(str(e.exception), "Argument property should be string")

        self.assertIn(self.session_id, self.cache.CACHE)
        self.assertDictEqual(
            self.cache.CACHE[self.session_id][Cache.session_details],
            self.session_details,
        )

    def test_get_cache_invalid_args(self):
        with self.assertRaises(Exception) as e:
            self.cache.get_cache(123, 123)
        self.assertEqual(str(e.exception), "Argument session_id should be string")

        with self.assertRaises(Exception) as e:
            self.cache.get_cache(self.session_id, 123)
        self.assertEqual(str(e.exception), "Argument property should be string")

    @patch.object(Cache, "cleanup_cache")
    def test_get_cache_success(self, mock_cleanup_cache):
        session_details = self.cache.get_cache(self.session_id, Cache.session_details)
        self.assertDictEqual(session_details, self.session_details)
        mock_cleanup_cache.assert_called()

    @patch("percy.cache.Cache.CACHE_TIMEOUT", 1)
    def test_cleanup_cache(self):
        cache_timeout = self.cache.CACHE_TIMEOUT
        time.sleep(cache_timeout + 1)
        self.assertIn(self.session_id, self.cache.CACHE)
        self.assertIn("key-1", self.cache.CACHE[self.session_id])
        self.cache.cleanup_cache()
        self.assertIn(self.session_id, self.cache.CACHE)
        self.assertIn("session_details", self.cache.CACHE[self.session_id])
        self.assertNotIn("key-1", self.cache.CACHE[self.session_id])



# pylint: disable=protected-access
import json
from percy.cache import Cache


class PageMetaData:
    def __init__(self, page):
        self.page = page

    def __fetch_guid(self, obj):
        return obj._impl_obj._guid

    @property
    def framework(self):
        return "playwright"

    @property
    def page_guid(self):
        return self.__fetch_guid(self.page)

    @property
    def frame_guid(self):
        return self.__fetch_guid(self.page.main_frame)

    @property
    def browser_guid(self):
        return self.__fetch_guid(self.page.context.browser)

    @property
    def session_details(self):
        session_details = Cache.get_cache(self.browser_guid, Cache.session_details)
        if session_details is None:
            session_details = json.loads(
                self.page.evaluate(
                    "_ => {}", 'browserstack_executor: {"action": "getSessionDetails"}'
                )
            )
            Cache.set_cache(self.browser_guid, Cache.session_details, session_details)
            return session_details
        return session_details

    @property
    def automate_session_id(self):
        return self.session_details.get("hashed_id")



from setuptools import setup
from os import path
import percy

# read the README for long_description
cwd = path.abspath(path.dirname(__file__))
with open(path.join(cwd, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='percy-playwright',
    description='Python client for visual testing with Percy',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=percy.__version__,
    license='MIT',
    author='Perceptual Inc.',
    author_email='team@percy.io',
    url='https://github.com/percy/percy-playwright-python',
    keywords='percy visual testing',
    packages=['percy'],
    include_package_data=True,
    install_requires=[
        'playwright>=1.28.0',
        'requests==2.*'
    ],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    test_suite='tests',
    tests_require=['playwright', 'httpretty'],
    zip_safe=False
)
